/******************************************************************************
 * i2c-test.c: the I2C test
 *
 * Copyright (C) 2011, MyVoice
 * All Rights reserved.
 *
 * History
 * 2011.06.30	ver 1.00	Preliminary version, first release
 *
 * This is the test for the EEPROM connected to the I2C1 interface.
 * It reads and writes a data pattern from the EEPROM.
 *
 * Software License Agreement
 *
 * The software is owned by MyVoice CAD/CAM Services and/or its suppliers, and is
 * protected under applicable copyright laws.  All rights are reserved.  Any
 * use in violation of the foregoing restrictions may subject the user to criminal
 * sanctions under applicable laws, as well as to civil liability for the breach
 * of the terms and conditions of this license.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS".  NO WARRANTIES, WHETHER EXPRESS, IMPLIED
 * OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE.
 * USE OF THIS SOFTWARE FOR COMMERCIAL DEVELOPMENT AND/OR EDUCATION IS SUBJECT
 * TO A CURRENT END USER LICENSE AGREEMENT (COMMERCIAL OR EDUCATIONAL) WITH
 * MYVOICE CAD/CAM SERVICES.
 *
 */
#include "LPC17xx.h"			/* LPC17xx Peripheral Registers */

#include "i2c.h"
#include "small_printf.h"
#include "tests.h"


#define MC24LC08_ADDR 0xa0		/* I2C address of the 24LC08 EEPROM */
#define I2CBUFSIZE    8			/* Use 8 byte buffers to store data */

/*
 * Test pattern
 */
#define DAT0	0x55
#define DAT1	0xaa
#define DAT2	0xa5
#define DAT3	0x5a

/*
 * i2cResetIO - reset the I2C lines back to GPIO after the test.
 *
 * Normally this is not needed but for the pin-test to pass we need the pins
 * in a defined GPIO state (in normal mode, not Open Drain).
 */
static void i2cResetIO(void)
{
	LPC_PINCON->PINSEL0     &= ~0b1111;
	LPC_PINCON->PINMODE_OD0 &= ~0b1111;
}

/*
 * i2cTest - try to read from the 24lc08 EEPROM
 *
 * Read 4 bytes from the EEPROM
 * If these bytes do not contain the test pattern (true for a new EEPROM)
 * we then write the test pattern and read again.
 */
void i2cTest (void)
{
	uint8_t I2CWriteBuffer[I2CBUFSIZE];
	uint8_t I2CReadBuffer[I2CBUFSIZE];
	uint8_t I2CWriteLength;
	uint8_t I2CReadLength;

	uint32_t i;
	uint32_t result;

	printf("I2C Test: read EEPROM and check for data pattern,\n\r write and check pattern if not found\n\r");

	if ( I2CInit( (uint32_t)I2CMASTER ) == FALSE )	/* initialize I2c */
	{
		errors++;
		printf("\tError: failed to initialize I2C\n\r");
		i2cResetIO();
		return;
	}

	/* In order to start the I2CEngine,  all the parameters
	 * must be set in advance, including I2CWriteLength, I2CReadLength,
	 * I2CCmd, and the I2cMasterBuffer which contains the stream
	 * command/data to the I2c slave device.
	 * (1) If it's a I2C write only, the number of bytes to be written is
	 *     I2CWriteLength, I2CReadLength is zero, the content will be filled
	 *     in the I2CMasterBuffer.
	 * (2) If it's a I2C read only, the number of bytes to be read is
	 *     I2CReadLength, I2CWriteLength is 0, the read value will be filled
	 *     in the I2CMasterBuffer.
	 * (3) If it's a I2C Write/Read with repeated start, specify the
	 *     I2CWriteLength, fill the content of bytes to be written in
	 *     I2CMasterBuffer, specify the I2CReadLength, after the repeated
	 *     start and the device address with RD bit set, the content of the
	 *     reading will be filled in I2CMasterBuffer index at
	 *     I2CMasterBuffer[I2CWriteLength+2].
	 *
	 * e.g. Start, DevAddr(W), WRByte1...WRByteN, Repeated-Start, DevAddr(R),
	 * RDByte1...RDByteN Stop. The content of the reading will be filled
	 * after (I2CWriteLength + two devaddr) bytes.
	 */

	/*
	 * Clear the I2C read buffer,
	 * then create a Write/Read transaction to read 4 bytes from address 0 in EEPROM
	 */
	for ( i = 0; i < I2CBUFSIZE; i++ )
	{
		I2CReadBuffer[i] = 0x00;
	}
	/* Write SLA(W), address, SLA(R), and read 4 bytes back. */

	I2CWriteLength = 2;
	I2CReadLength = 4;
	I2CWriteBuffer[0] = MC24LC08_ADDR;
	I2CWriteBuffer[1] = 0x00;		/* address */
	I2CWriteBuffer[2] = MC24LC08_ADDR | RD_BIT;

	result = I2CEngine(I2CWriteBuffer, &I2CWriteLength, I2CReadBuffer, &I2CReadLength);

	if(result == I2CSTATE_ACK)
	{
		if((I2CReadBuffer[0] == DAT0) &&
		   (I2CReadBuffer[1] == DAT1) &&
		   (I2CReadBuffer[2] == DAT2) &&
		   (I2CReadBuffer[3] == DAT3))
		{
			printf("\tI2C Read Transaction passed, data OK\n\r");
		} else {
			/*
			 * We had a valid read but no valid data.
			 *
			 */
			I2CWriteLength = 6;
			I2CReadLength = 0;
			I2CWriteBuffer[0] = MC24LC08_ADDR;
			I2CWriteBuffer[1] = 0x00;		/* address */
			I2CWriteBuffer[2] = DAT0;		/* Data0 */
			I2CWriteBuffer[3] = DAT1;		/* Data1 */
			I2CWriteBuffer[4] = DAT2;		/* Data0 */
			I2CWriteBuffer[5] = DAT3;		/* Data1 */
			I2CEngine(I2CWriteBuffer, &I2CWriteLength, I2CReadBuffer, &I2CReadLength);

			/*
			 * and read again
			 *
			 * Clear buffer, create transaction etc...
			 *
			 * Since this read takes place after a write, the EEPROM may still be busy
			 * so we try to read 10 times before signalling a real error.
			 */
			for ( i = 0; i < I2CBUFSIZE; i++ )
			{
				I2CReadBuffer[i] = 0x00;
			}

			I2CWriteLength = 2;
			I2CReadLength = 4;
			I2CWriteBuffer[0] = MC24LC08_ADDR;
			I2CWriteBuffer[1] = 0x00;		/* address */
			I2CWriteBuffer[2] = MC24LC08_ADDR | RD_BIT;

			i=0;
			do
			{
				delay(5);
				result = I2CEngine(I2CWriteBuffer, &I2CWriteLength, I2CReadBuffer, &I2CReadLength);
				i++;
			} while ((result == I2CSTATE_SLA_NACK) && (i < 10));

			if(result == I2CSTATE_ACK)
			{
				if((I2CReadBuffer[0] == DAT0) &&
				   (I2CReadBuffer[1] == DAT1) &&
				   (I2CReadBuffer[2] == DAT2) &&
				   (I2CReadBuffer[3] == DAT3))
				{
					printf("\tI2C Write and Read Transactions passed, data OK\n\r");
				} else {
					printf("\tI2C Write and Read Transactions passed, data not OK\n\r");
					errors++;
				}
			} else {
				printf("\tI2C Transaction error, result is: %d\n\r", result);
				errors++;
			}
		}
	}
	else
	{
		printf("\tI2C Read Transaction error, result is: %d\n\r", result);
		errors++;
	}

	i2cResetIO();
}

/******************************************************************************
**                            End Of File
******************************************************************************/
